@@ -63,7 +63,29 @@
 #endif
 
 #if !OPENCONNECT_CHECK_VER(2,1)
-#define openconnect_set_stoken_mode(...) -EOPNOTSUPP
+#define __openconnect_set_token_mode(...) -EOPNOTSUPP
+#elif !OPENCONNECT_CHECK_VER(2,2)
+#define __openconnect_set_token_mode(vpninfo, mode, secret) openconnect_set_stoken_mode(vpninfo, 1, secret)
+#else
+#define __openconnect_set_token_mode openconnect_set_token_mode
+#endif
+
+#if OPENCONNECT_CHECK_VER(3,0)
+#define NEWGROUP_SUPPORTED		1
+#define AUTHGROUP_OPT(form)		(void *)(form)->authgroup_opt
+#define AUTHGROUP_SELECTION(form)	(form)->authgroup_selection
+#define FORMCHOICE(sopt, i)		((sopt)->choices[i])
+#define IGNORE_OPT(opt)			((opt)->flags & OC_FORM_OPT_IGNORE)
+#else
+#define NEWGROUP_SUPPORTED		0
+#define AUTHGROUP_OPT(form)		NULL
+#define AUTHGROUP_SELECTION(form)	0
+#define FORMCHOICE(sopt, i)		(&(sopt)->choices[i])
+#define IGNORE_OPT(opt)			0
+#define OC_FORM_RESULT_ERR		-1
+#define OC_FORM_RESULT_OK		0
+#define OC_FORM_RESULT_CANCELLED	1
+#define OC_FORM_RESULT_NEWGROUP		2
 #endif
 
 #ifdef OPENCONNECT_OPENSSL
@@ -179,6 +201,9 @@ typedef struct auth_ui_data {
 	GCond *form_shown_changed;
 	gboolean form_shown;
 
+	gboolean newgroup;
+	gboolean group_set;
+
 	GCond *cert_response_changed;
 	enum certificate_response cert_response;
 
@@ -219,8 +244,8 @@ static void ssl_box_add_error(auth_ui_data *ui_data, const char *msg)
 #endif
 	gtk_box_pack_start(GTK_BOX(ui_data->ssl_box), hbox, FALSE, FALSE, 0);
 
-	image = gtk_image_new_from_stock(GTK_STOCK_DIALOG_ERROR,
-					 GTK_ICON_SIZE_DIALOG);
+	image = gtk_image_new_from_icon_name("dialog-error",
+					     GTK_ICON_SIZE_DIALOG);
 	gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
 
 	text = gtk_label_new(msg);
@@ -240,8 +265,8 @@ static void ssl_box_add_notice(auth_ui_data *ui_data, const char *msg)
 #endif
 	gtk_box_pack_start(GTK_BOX(ui_data->ssl_box), hbox, FALSE, FALSE, 0);
 
-	image = gtk_image_new_from_stock(GTK_STOCK_DIALOG_WARNING,
-					 GTK_ICON_SIZE_DIALOG);
+	image = gtk_image_new_from_icon_name("dialog-warning",
+					     GTK_ICON_SIZE_DIALOG);
 	gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
 
 	text = gtk_label_new(msg);
@@ -285,6 +310,7 @@ typedef struct ui_fragment_data {
 #endif
 	struct oc_form_opt *opt;
 	char *entry_text;
+	int initial_selection;
 	int grab_focus;
 } ui_fragment_data;
 
@@ -339,6 +365,13 @@ static void do_override_label(ui_fragment_data *data, struct oc_choice *choice)
 	gtk_label_set_text(GTK_LABEL(data->widget), new_label);
 
 }
+
+static gboolean do_newgroup(GtkDialog *dialog)
+{
+	gtk_dialog_response(dialog, AUTH_DIALOG_RESPONSE_LOGIN);
+	return FALSE;
+}
+
 static void combo_changed(GtkComboBox *combo, ui_fragment_data *data)
 {
 	struct oc_form_opt_select *sopt = (void *)data->opt;
@@ -346,10 +379,16 @@ static void combo_changed(GtkComboBox *combo, ui_fragment_data *data)
 	if (entry < 0)
 		return;
 
-	data->entry_text = sopt->choices[entry].name;
+	data->entry_text = FORMCHOICE(sopt, entry)->name;
+
+	if (NEWGROUP_SUPPORTED && entry != data->initial_selection) {
+		data->ui_data->newgroup = TRUE;
+		g_idle_add ((GSourceFunc)do_newgroup, data->ui_data->dialog);
+		return;
+	}
 
 	g_queue_foreach(data->ui_data->form_entries, (GFunc)do_override_label,
-			&sopt->choices[entry]);
+			FORMCHOICE(sopt, entry));
 }
 
 #ifdef OPENCONNECT_OPENSSL
@@ -430,7 +469,7 @@ static gboolean ui_add_select (ui_fragment_data *data)
 	auth_ui_data *ui_data = _ui_data; /* FIXME global */
 	GtkWidget *hbox, *text, *combo;
 	struct oc_form_opt_select *sopt = (void *)data->opt;
-	int i;
+	int i, user_selection = -1;
 
 #if GTK_CHECK_VERSION(3,1,6)
 	hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
@@ -444,20 +483,21 @@ static gboolean ui_add_select (ui_fragment_data *data)
 
 	combo = gtk_combo_box_text_new();
 	gtk_box_pack_end(GTK_BOX(hbox), combo, FALSE, FALSE, 0);
+
 	for (i = 0; i < sopt->nr_choices; i++) {
-		gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(combo), sopt->choices[i].label);
+		gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(combo), FORMCHOICE(sopt, i)->label);
 		if (data->entry_text &&
-		    !strcmp(data->entry_text, sopt->choices[i].name)) {
-			gtk_combo_box_set_active(GTK_COMBO_BOX(combo), i);
-			g_free(data->entry_text);
-			data->entry_text = sopt->choices[i].name;
-		}
-	}
-	if (gtk_combo_box_get_active(GTK_COMBO_BOX(combo)) < 0) {
-		gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0); 
-		data->entry_text = sopt->choices[0].name;
+		    !strcmp(data->entry_text, FORMCHOICE(sopt, i)->name))
+			user_selection = i;
 	}
 
+	i = data->initial_selection != -1 ? data->initial_selection :
+	    user_selection != -1 ? user_selection : 0;
+	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), i);
+	g_free(data->entry_text);
+	data->entry_text = FORMCHOICE(sopt, i)->name;
+	data->initial_selection = i;
+
 	if (g_queue_peek_tail(ui_data->form_entries) == data)
 		gtk_widget_grab_focus (combo);
 	g_signal_connect(G_OBJECT(combo), "changed", G_CALLBACK(combo_changed), data);
@@ -666,7 +706,8 @@ static gboolean ui_form (struct oc_auth_form *form)
 	for (opt = form->opts; opt; opt = opt->next) {
 		ui_fragment_data *data;
 
-		if (opt->type == OC_FORM_OPT_HIDDEN)
+		if (opt->type == OC_FORM_OPT_HIDDEN ||
+		    IGNORE_OPT(opt))
 			continue;
 
 		data = g_slice_new0 (ui_fragment_data);
@@ -678,10 +719,12 @@ static gboolean ui_form (struct oc_auth_form *form)
 			g_mutex_lock (ui_data->form_mutex);
 			g_queue_push_head(ui_data->form_entries, data);
 			g_mutex_unlock (ui_data->form_mutex);
-			if (opt->type != OC_FORM_OPT_PASSWORD)
+			if (opt->type != OC_FORM_OPT_PASSWORD) {
 				data->entry_text = g_strdup (find_form_answer(ui_data->secrets,
 									      form, opt));
-			else {
+				if (!data->entry_text)
+					data->entry_text = g_strdup (opt->value);
+			} else {
 				data->find_request = gnome_keyring_find_password(
 						OPENCONNECT_SCHEMA,
 						got_keyring_pw,
@@ -703,6 +746,11 @@ static gboolean ui_form (struct oc_auth_form *form)
 			data->entry_text = g_strdup (find_form_answer(ui_data->secrets,
 								      form, opt));
 
+			if (opt == AUTHGROUP_OPT(form))
+				data->initial_selection = AUTHGROUP_SELECTION(form);
+			else
+				data->initial_selection = -1;
+
 			ui_add_select(data);
 		} else
 			g_slice_free (ui_fragment_data, data);
@@ -711,11 +759,51 @@ static gboolean ui_form (struct oc_auth_form *form)
 	return ui_show(ui_data);
 }
 
+/* If our stored group_list selection differs from the server default, send a
+   NEWGROUP request to try to change it before rendering the form */
+
+static gboolean set_initial_authgroup (auth_ui_data *ui_data, struct oc_auth_form *form)
+{
+	struct oc_form_opt *opt;
+
+	if (!NEWGROUP_SUPPORTED || ui_data->group_set || !AUTHGROUP_OPT(form))
+		return FALSE;
+	ui_data->group_set = TRUE;
+
+	for (opt = form->opts; opt; opt = opt->next) {
+		int i;
+		char *saved_group;
+		struct oc_form_opt_select *sopt;
+
+		if (opt != AUTHGROUP_OPT(form))
+			continue;
+
+		saved_group = find_form_answer(ui_data->secrets, form, opt);
+		if (!saved_group)
+			return FALSE;
+
+		sopt = (struct oc_form_opt_select *)opt;
+		for (i = 0; i < sopt->nr_choices; i++) {
+			struct oc_choice *ch = FORMCHOICE(sopt, i);
+			if (!strcmp(saved_group, ch->name) && i != AUTHGROUP_SELECTION(form)) {
+				free(opt->value);
+				opt->value = g_strdup(saved_group);
+				return TRUE;
+			}
+		}
+	}
+	return FALSE;
+}
+
 static int nm_process_auth_form (void *cbdata, struct oc_auth_form *form)
 {
 	auth_ui_data *ui_data = cbdata;
 	int response;
 
+	if (set_initial_authgroup(ui_data, form))
+		return OC_FORM_RESULT_NEWGROUP;
+
+	ui_data->newgroup = FALSE;
 	g_idle_add((GSourceFunc)ui_form, form);
 
 	g_mutex_lock(ui_data->form_mutex);
@@ -776,10 +864,13 @@ static int nm_process_auth_form (void *cbdata, struct oc_auth_form *form)
 	ui_data->form_grabbed = 0;
 	g_mutex_unlock(ui_data->form_mutex);
 	
-	/* -1 = cancel,
-	 *  0 = failure,
-	 *  1 = success */
-	return (response == AUTH_DIALOG_RESPONSE_LOGIN ? 0 : 1);
+	if (response == AUTH_DIALOG_RESPONSE_LOGIN) {
+		if (ui_data->newgroup)
+			return OC_FORM_RESULT_NEWGROUP;
+		else
+			return OC_FORM_RESULT_OK;
+	} else
+		return OC_FORM_RESULT_CANCELLED;
 
 }
 
@@ -1039,8 +1130,8 @@ static int get_config (GHashTable *options, GHashTable *secrets,
 	char *csd_wrapper;
 	char *pem_passphrase_fsid;
 	char *cafile;
-	char *stoken_source;
-	char *stoken_string;
+	char *token_mode;
+	char *token_secret;
 
 	hostname = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_GATEWAY);
 	if (!hostname) {
@@ -1111,18 +1202,20 @@ static int get_config (GHashTable *options, GHashTable *secrets,
 	if (pem_passphrase_fsid && cert && !strcmp(pem_passphrase_fsid, "yes"))
 		openconnect_passphrase_from_fsid(vpninfo);
 
-	stoken_source = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_STOKEN_SOURCE);
-	stoken_string = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_STOKEN_STRING);
-	if (stoken_source) {
+	token_mode = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_TOKEN_MODE);
+	token_secret = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_TOKEN_SECRET);
+	if (token_mode) {
 		int ret = 0;
 
-		if (!strcmp(stoken_source, "manual") && stoken_string)
-			ret = openconnect_set_stoken_mode(vpninfo, 1, stoken_string);
-		else if (!strcmp(stoken_source, "stokenrc"))
-			ret = openconnect_set_stoken_mode(vpninfo, 1, NULL);
+		if (!strcmp(token_mode, "manual") && token_secret)
+			ret = __openconnect_set_token_mode(vpninfo, OC_TOKEN_MODE_STOKEN, token_secret);
+		else if (!strcmp(token_mode, "stokenrc"))
+			ret = __openconnect_set_token_mode(vpninfo, OC_TOKEN_MODE_STOKEN, NULL);
+		else if (!strcmp(token_mode, "totp") && token_secret)
+			ret = __openconnect_set_token_mode(vpninfo, OC_TOKEN_MODE_TOTP, token_secret);
 
 		if (ret)
-			fprintf(stderr, "Failed to initialize stoken: %d\n", ret);
+			fprintf(stderr, "Failed to initialize software token: %d\n", ret);
 	}
 
 	return 0;
@@ -1288,9 +1381,9 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 	if (ui_data->cookie_retval < 0) {
 		/* error while getting cookie */
 		if (ui_data->last_notice_icon) {
-			gtk_image_set_from_stock(GTK_IMAGE (ui_data->last_notice_icon),
-						 GTK_STOCK_DIALOG_ERROR,
-						 GTK_ICON_SIZE_DIALOG);
+			gtk_image_set_from_icon_name(GTK_IMAGE (ui_data->last_notice_icon),
+						     "dialog-error",
+						     GTK_ICON_SIZE_DIALOG);
 			gtk_widget_show_all(ui_data->ssl_box);
 			gtk_widget_set_sensitive(ui_data->cancel_button, FALSE);
 		}
@@ -1469,14 +1562,14 @@ static void login_clicked (GtkButton *btn, auth_ui_data *ui_data)
 static void build_main_dialog(auth_ui_data *ui_data)
 {
 	char *title;
-	GtkWidget *vbox, *hbox, *label, *frame, *image, *frame_box;
+	GtkWidget *vbox, *hbox, *label, *frame, *frame_box;
 	GtkWidget *exp, *scrolled, *view, *autocon, *save_pass;
 
-	gtk_window_set_default_icon_name(GTK_STOCK_DIALOG_AUTHENTICATION);
+	gtk_window_set_default_icon_name("dialog-password");
 
 	title = get_title(ui_data->vpn_name);
 	ui_data->dialog = gtk_dialog_new_with_buttons(title, NULL, GTK_DIALOG_MODAL,
-						      GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
+						      _("_Close"), GTK_RESPONSE_CLOSE,
 						      NULL);
 	g_signal_connect (ui_data->dialog, "response", G_CALLBACK(dialog_response), ui_data);
 	gtk_window_set_default_size(GTK_WINDOW(ui_data->dialog), 350, 300);
@@ -1512,10 +1605,8 @@ static void build_main_dialog(auth_ui_data *ui_data)
 	                         G_CALLBACK(queue_connect_host), ui_data);
 	gtk_widget_show(ui_data->combo);
 
-	ui_data->connect_button = gtk_button_new();
+	ui_data->connect_button = gtk_button_new_with_mnemonic(_("C_onnect"));
 	gtk_box_pack_end(GTK_BOX(hbox), ui_data->connect_button, FALSE, FALSE, 0);
-	image = gtk_image_new_from_stock(GTK_STOCK_CONNECT, GTK_ICON_SIZE_BUTTON);
-	gtk_button_set_image (GTK_BUTTON(ui_data->connect_button), image);
 	gtk_widget_grab_focus(ui_data->connect_button);
 	g_signal_connect_swapped(ui_data->connect_button, "clicked",
 				 G_CALLBACK(queue_connect_host), ui_data);
@@ -1568,14 +1659,12 @@ static void build_main_dialog(auth_ui_data *ui_data)
 	gtk_widget_show(hbox);
 
 	ui_data->login_button = gtk_button_new_with_mnemonic(_("_Login"));
-	image = gtk_image_new_from_stock(GTK_STOCK_APPLY, GTK_ICON_SIZE_BUTTON);
-	gtk_button_set_image (GTK_BUTTON(ui_data->login_button), image);
 	gtk_box_pack_end(GTK_BOX(hbox), ui_data->login_button, FALSE, FALSE, 0);
 	g_signal_connect (ui_data->login_button, "clicked", G_CALLBACK(login_clicked), ui_data);
 	gtk_widget_set_sensitive (ui_data->login_button, FALSE);
 	gtk_widget_show(ui_data->login_button);
 
-	ui_data->cancel_button = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
+	ui_data->cancel_button = gtk_button_new_with_mnemonic(_("_Cancel"));
 	gtk_box_pack_end(GTK_BOX(hbox), ui_data->cancel_button, FALSE, FALSE, 0);
 	g_signal_connect (ui_data->cancel_button, "clicked", G_CALLBACK(cancel_clicked), ui_data);
 	gtk_widget_set_sensitive (ui_data->cancel_button, FALSE);
