@@ -45,6 +45,9 @@
 #if !OPENCONNECT_CHECK_VER(2,1)
 #define openconnect_has_stoken_support() 0
 #endif
+#if !OPENCONNECT_CHECK_VER(2,2)
+#define openconnect_has_oath_support() 0
+#endif
 
 #define NM_VPN_API_SUBJECT_TO_CHANGE
 
@@ -175,15 +178,15 @@ import (NMVpnPluginUiInterface *iface, const char *path, GError **error)
 	if (bval)
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID, "yes");
 
-	/* Soft token source */
+	/* Soft token mode */
 	buf = g_key_file_get_string (keyfile, "openconnect", "StokenSource", NULL);
 	if (buf)
-		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_STOKEN_SOURCE, buf);
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_MODE, buf);
 
-	/* Soft token string */
+	/* Soft token secret */
 	buf = g_key_file_get_string (keyfile, "openconnect", "StokenString", NULL);
 	if (buf)
-		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_STOKEN_STRING, buf);
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_SECRET, buf);
 
 	return connection;
 }
@@ -205,8 +208,8 @@ export (NMVpnPluginUiInterface *iface,
 	const char *usercert = NULL;
 	const char *privkey = NULL;
 	gboolean pem_passphrase_fsid = FALSE;
-	const char *stoken_source = NULL;
-	const char *stoken_string = NULL;
+	const char *token_mode = NULL;
+	const char *token_secret = NULL;
 	gboolean success = FALSE;
 	FILE *f;
 
@@ -256,13 +259,13 @@ export (NMVpnPluginUiInterface *iface,
 	if (value && !strcmp (value, "yes"))
 		pem_passphrase_fsid = TRUE;
 
-	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_STOKEN_SOURCE);
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_MODE);
 	if (value && strlen (value))
-		stoken_source = value;
+		token_mode = value;
 
-	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_STOKEN_STRING);
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_SECRET);
 	if (value && strlen (value))
-		stoken_string = value;
+		token_secret = value;
 
 	fprintf (f,
 		 "[openconnect]\n"
@@ -286,8 +289,8 @@ export (NMVpnPluginUiInterface *iface,
 		 /* User Certificate */      usercert,
 		 /* Private Key */           privkey,
 		 /* FSID */                  pem_passphrase_fsid ? "1" : "0",
-		 /* Soft token source */     stoken_source ? stoken_source : "",
-		 /* Soft token string */     stoken_string ? stoken_string : "");
+		 /* Soft token mode */       token_mode ? token_mode : "",
+		 /* Soft token secret */     token_secret ? token_secret : "");
 
 	success = TRUE;
 
@@ -372,47 +375,90 @@ stuff_changed_cb (GtkWidget *widget, gpointer user_data)
 }
 
 static gboolean
-init_stoken_ui (OpenconnectPluginUiWidget *self,
+init_token_mode_options (GtkComboBox *token_mode)
+{
+	GtkListStore *token_mode_list = GTK_LIST_STORE (gtk_combo_box_get_model (token_mode));
+	GtkTreeModel *model = GTK_TREE_MODEL (token_mode_list);
+	GtkTreeIter iter;
+	gboolean iter_valid;
+	int valid_rows = 0;
+
+	if (!gtk_tree_model_get_iter_first (model, &iter))
+		return FALSE;
+	do {
+		char *token_type;
+
+		gtk_tree_model_get (model, &iter, 2, &token_type, -1);
+		if (!strcmp (token_type, "stoken") && !openconnect_has_stoken_support ())
+			iter_valid = gtk_list_store_remove (token_mode_list, &iter);
+		else if (!strcmp (token_type, "totp") && !openconnect_has_oath_support ())
+			iter_valid = gtk_list_store_remove (token_mode_list, &iter);
+		else {
+			iter_valid = gtk_tree_model_iter_next (model, &iter);
+			valid_rows++;
+		}
+		g_free (token_type);
+	} while (iter_valid);
+
+	/* if the only option is "Disabled", don't show the token section at all */
+	return valid_rows > 1;
+}
+
+static gboolean
+init_token_ui (OpenconnectPluginUiWidget *self,
 				OpenconnectPluginUiWidgetPrivate *priv,
 				NMSettingVPN *s_vpn)
 {
 	GtkWidget *widget;
+	GtkComboBox *token_mode;
 	GtkTextBuffer *buffer;
 	const char *value;
 
-	/* don't advertise stoken properties if we can't use them anyway */
-	if (!openconnect_has_stoken_support ())
+	token_mode = GTK_COMBO_BOX (gtk_builder_get_object (priv->builder, "token_mode"));
+	if (!token_mode)
+		return FALSE;
+	if (!init_token_mode_options (token_mode))
 		return TRUE;
 
-	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "stoken_vbox"));
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "token_vbox"));
 	if (!widget)
 		return FALSE;
 	gtk_box_pack_start (GTK_BOX (priv->widget), widget, FALSE, FALSE, 0);
 
-	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "stoken_source"));
-	if (!widget)
-		return FALSE;
 	if (s_vpn) {
-		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_STOKEN_SOURCE);
-		if (value) {
-			if (!strcmp (value, "stokenrc"))
-				gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 1);
-			else if (!strcmp (value, "manual"))
-				gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 2);
-			else
-				gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
-		}
-	}
-	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);
+		GtkTreeModel *model = gtk_combo_box_get_model (token_mode);
+		int active_option = 0;
 
-	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "stoken_string"));
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_MODE);
+		if (value) {
+			int i;
+			GtkTreeIter iter;
+
+			if (!gtk_tree_model_get_iter_first (model, &iter))
+				return FALSE;
+			for (i = 0; ; i++) {
+				char *pref_value;
+
+				gtk_tree_model_get (model, &iter, 1, &pref_value, -1);
+				if (!strcmp (value, pref_value))
+					active_option = i;
+				g_free (pref_value);
+				if (!gtk_tree_model_iter_next (model, &iter))
+					break;
+			}
+		}
+		gtk_combo_box_set_active (token_mode, active_option);
+	}
+	g_signal_connect (G_OBJECT (token_mode), "changed", G_CALLBACK (stuff_changed_cb), self);
+
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "token_secret"));
 	if (!widget)
 		return FALSE;
 	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widget));
 	if (!buffer)
 		return FALSE;
 	if (s_vpn) {
-		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_STOKEN_STRING);
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_SECRET);
 		if (value)
 			gtk_text_buffer_set_text (buffer, value, -1);
 	}
@@ -485,7 +531,7 @@ init_plugin_ui (OpenconnectPluginUiWidget *self, NMConnection *connection, GErro
 	}
 	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);
 
-	if (init_stoken_ui (self, priv, s_vpn) == FALSE)
+	if (init_token_ui (self, priv, s_vpn) == FALSE)
 		return FALSE;
 
 	tls_pw_init_auth_widget (priv->builder, priv->group, s_vpn, stuff_changed_cb, self);
@@ -512,8 +558,9 @@ update_connection (NMVpnPluginUiWidgetInterface *iface,
 	NMSettingVPN *s_vpn;
 	GtkWidget *widget;
 	char *str;
-	gint idx;
-	gboolean stoken_string_editable = FALSE;
+	GtkTreeModel *model;
+	GtkTreeIter iter;
+	gboolean token_secret_editable = FALSE;
 	GtkTextIter iter_start, iter_end;
 	GtkTextBuffer *buffer;
 	const char *auth_type = NULL;
@@ -544,29 +591,19 @@ update_connection (NMVpnPluginUiWidgetInterface *iface,
 	if (str && strlen (str))
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_CSD_WRAPPER, str);
 
-	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "stoken_source"));
-	idx = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
-	str = NULL;
-	switch (idx) {
-	case 0:
-		str = "disabled";
-		break;
-	case 1:
-		str = "stokenrc";
-		break;
-	case 2:
-		str = "manual";
-		stoken_string_editable = TRUE;
-		break;
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "token_mode"));
+	model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));
+	if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter)) {
+		gtk_tree_model_get (model, &iter, 1, &str, 3, &token_secret_editable, -1);
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_MODE, str);
+		g_free(str);
 	}
-	if (str)
-		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_STOKEN_SOURCE, str);
 
-	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "stoken_string_label"));
-	gtk_widget_set_sensitive (widget, stoken_string_editable);
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "token_secret_label"));
+	gtk_widget_set_sensitive (widget, token_secret_editable);
 
-	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "stoken_string"));
-	gtk_widget_set_sensitive (widget, stoken_string_editable);
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "token_secret"));
+	gtk_widget_set_sensitive (widget, token_secret_editable);
 
 	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widget));
 	gtk_text_buffer_get_start_iter (buffer, &iter_start);
@@ -577,12 +614,12 @@ update_connection (NMVpnPluginUiWidgetInterface *iface,
 
 		/* zap invalid characters */
 		for (; *src; src++)
-			if ((*src >= '0' && *src <= '9') || *src == '-')
+			if (*src >= ' ' && *src <= '~')
 				*(dst++) = *src;
 		*dst = 0;
 
 		if (strlen (str))
-			nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_STOKEN_STRING, str);
+			nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_SECRET, str);
 	}
 
 	if (!check_validity (self, error))
